#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#ifdef WINDOWS
#include <windows.h>
#include <winsock.h>
#endif

#include "lzw.h"
#include "bmp.h"
#include "fileio.h"
#include "mandel_calc.h"

/*

This code is copyright 2004-2005 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

int ptr=0,lptr=0,lr=0;
unsigned char myString[1024];

int send_request(int socketid, char *message)
{
fd_set writeset;
int t;

  FD_ZERO(&writeset);
  FD_SET(socketid,&writeset);

  if (((t=select(socketid+1,NULL,&writeset,NULL,NULL))==-1) && (errno!=EINTR))
  {
    printf("Problem with select\n");
    return -1;
  }
    else
  { send(socketid,message,strlen(message),0); }

  return 0;
}

int parse_command(char *message)
{
int s;

  if (strlen(message)<4)
  {
    return -1;
  }

  s=3;
  while(message[s]!=' ' && message[s]!='\r' && message[s]!='\n' && message[s]!=0)
  {
    s++;
  }

  message[s]=0;

  return 0;
}

int parse_params(char *buffer, struct mandelbrot_t *mandelbrot)
{
int r,t;
int params[8];

  r=0;
  for (t=0; t<8; t++)
  {
    params[t]=r;
    while(buffer[r]!=',' && buffer[r]!=0) r++;
    buffer[r++]=0;
  }

  mandelbrot->width=atoi(buffer+params[0]);
  mandelbrot->height=atoi(buffer+params[1]);
  mandelbrot->real_start=atof(buffer+params[2]);
  mandelbrot->real_end=atof(buffer+params[3]);
  mandelbrot->imaginary_start=atof(buffer+params[4]);
  mandelbrot->imaginary_end=atof(buffer+params[5]);
  mandelbrot->count=atoi(buffer+params[6]);

  printf("     Real start: %f\n",mandelbrot->real_start);
  printf("       Real end: %f\n",mandelbrot->real_end);
  printf("Imaginary start: %f\n",mandelbrot->imaginary_start);
  printf("  Imaginary end: %f\n",mandelbrot->imaginary_end);
  printf("          Width: %d\n",mandelbrot->width);
  printf("         Height: %d\n",mandelbrot->height);
  printf("          Count: %d\n",mandelbrot->count);
  printf("\n");

  return 0;
}

int read_buffer(int socketid, unsigned char *outString)
{
fd_set readset;
struct timeval tv;
int r,t=0;
int tries;

  tries=0;
  outString[0]=0;

  while(1)
  {
    errno=0;
    if (lr>=lptr)
    {
      tv.tv_sec=10;
      tv.tv_usec=0;

      FD_ZERO(&readset);
      FD_SET(socketid,&readset);
      if (((t=select(socketid+1,&readset,NULL,NULL,&tv))==-1) && (errno!=EINTR))
      { return -1; }

      if (FD_ISSET(socketid,&readset))
      {  }
        else
      {
        tries++;
        if (tries==5)
        {
          printf("Problem reading from socket.\n");
          return -1;
        }
        continue;
      }

      tries=0;
      lptr=recv(socketid,(char *)myString,512-ptr,0);

      lr=0;
      if (lptr>=0) myString[lptr]=0;

      if (lptr==0) return -1;
      if (lptr<0)
      {
        ptr=0;
        lptr=0;
        return -1;
      }
    }

    for (r=lr; r<lptr; r++)
    {
      if ((myString[r]=='\n' || myString[r]=='\r') || myString[r]==0)
      {
        t=-1;
        break;
      }

      if (myString[r]==127 || myString[r]==8)
      { if (ptr>0) ptr--; }
        else
      { if (myString[r]>=32 && myString[r]<255) outString[ptr++]=myString[r]; }

      t=1;
    }

    lr=r+1;
    if ((t==-1 && ptr>0) || ptr>512)
    {
      outString[ptr]=0;
      ptr=0;
      if (myString[lr]=='\n' || myString[lr]=='\r')
      { lr++; }

      return 1;
    }
  }
}


int main(int argc, char *argv[])
{
FILE *out,*in; /* ,*socketfd; */
int socketid;
unsigned int *picture=0;
struct mandelbrot_t mandelbrot;
unsigned char buffer[1024];
unsigned char *command;
int t,r,l;
struct sockaddr_in serv_addr;
struct hostent *hp;
#ifdef WINDOWS
WSADATA wsaData;
WORD wVersionRequested=MAKEWORD(1,1);
int Win32isStupid;

  Win32isStupid=WSAStartup(wVersionRequested, &wsaData);
  if (Win32isStupid)  /* Shouldn't this always be true? :) */
  {
    printf("Winsock can't start.\n");
    exit(1);
  }

#endif

  printf("\nMandel client routines - January 9, 2005\n");
  printf("Copyright 2004-2005 - Michael Kohn <mike@mikekohn.net>\n\n");

  if (argc!=2)
  {
    printf("\nUsage: mandel_client <server address>\n");
    exit(0);
  }

  if ((hp=gethostbyname(argv[1]))==0)
  {
    printf("Host lookup failed %s.\n",argv[1]);
    exit(1);
  }

  /* bzero((char *) &serv_addr,sizeof(serv_addr)); */
  memset((char*)&serv_addr,0,sizeof(serv_addr)); 
  serv_addr.sin_family=AF_INET;
  serv_addr.sin_port=htons(5555);
  memcpy((char *)&serv_addr.sin_addr,hp->h_addr_list[0],hp->h_length);

  if ((socketid=socket(AF_INET,SOCK_STREAM,0))<0)
  {
    printf("Could not open socket.\n");
    exit(1);
  }

  if (connect(socketid,(struct sockaddr *) &serv_addr,sizeof(serv_addr))<0)
  {
    printf("Could not connect to server.\n");
    exit(1);
  }

  printf("Connected...\n");
  /* socketfd=fdopen(socketid,"r+"); */
  command=buffer+3;

  /* fgets(buffer,1023,socketfd); */
  read_buffer(socketid,buffer);
  printf("%s\n",buffer);

  /* fgets(buffer,1023,socketfd); */
  read_buffer(socketid,buffer);
  printf("%s\n",buffer);

  while(1)
  {
    send_request(socketid,"next\n");

    /* fgets(buffer,1023,socketfd); */
    read_buffer(socketid,buffer);
    printf("%s\n",buffer);

    if (parse_command(buffer)!=0) break;
    printf("Server command %s\n",command);
    fflush(stdout);

    if (strcasecmp(command,"error")==0)
    { break; }
      else
    if (strcasecmp(command,"done")==0)
    {
      printf("Finished.\n");
      break;
    }
      else
    if (strcasecmp(command,"compute")==0)
    {
      parse_params(buffer+11,&mandelbrot);
    }
      else
    {
      continue;
    }

    if (picture==0)
    { picture=(unsigned int *)malloc(mandelbrot.width*mandelbrot.height*sizeof(int)); }

    if (picture!=0)
    {
      mandel_calc(picture,&mandelbrot);

      out=fopen("image.gif","wb");
      if (out!=0)
      {
        lzw_compress(out,picture,mandelbrot.width,mandelbrot.height);
        fclose(out);
      }
        else
      {
        printf("Error opening image.gif for writing.\n");
        break;
      }
    }

    printf("*** UPLOADING ***\n");
    send_request(socketid,"upload\n");

    /* fgets(buffer,1023,socketfd); */
    read_buffer(socketid,buffer);
    printf("%s\n",buffer);

    if (parse_command(buffer)!=0) break;
    printf("Server command %s\n",command);

    if (strcasecmp(command,"error:")==0)
    { break; }
      else
    if (strcasecmp(command,"send")!=0)
    { break; }

    in=fopen("image.gif","rb");
    if (in==0)
    {
      printf("Error opening image.gif for reading.\n");
      break;
    }
    fseek(in,0,SEEK_END);
    l=ftell(in);
    fseek(in,0,SEEK_SET);

    buffer[0]=l>>24;
    buffer[1]=(l>>16)&255;
    buffer[2]=(l>>8)&255;
    buffer[3]=l&255;

    send(socketid,buffer,4,0);

    printf("Sent image size %d\n",l);
    fflush(stdout);

    while(l>0)
    {
      r=fread(buffer,1,1024,in);
      l=l-r;

      t=0;
      while(t!=r)
      {
        t=t+send(socketid,buffer+t,r-t,0);
      }
    }

    fclose(in);
  }

  free(picture);

  close(socketid);

  return 0;
}



