#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#endif
#include <errno.h>
#include <time.h>
#include <ctype.h>
#ifdef WINDOWS
#include <windows.h>
#endif

#include "globals.h"
#include "functions.h"

/*

This code is copyright 2004-2005 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

void safestrncpy(char *dest, char *source, int n)
{
int r;

  for (r=0; r<n; r++)
  { 
    dest[r]=source[r]; 
    if (source[r]==0) break;
  }  

  dest[r]=0;
}

int socketdie(int socketid)
{
int i;

  i=shutdown(socketid,SHUT_RDWR);
#ifdef WINDOWS
  closesocket(socketid);
#else
  close(socketid);
#endif
  return i;
}

char *strcasestr_m(char *hejstack, char *needle)
{
int hejlen,neelen;
int t,r;

  hejlen=strlen(hejstack);
  neelen=strlen(needle);
  r=hejlen-neelen;

  for(t=0; t<=r; t++)
  {
    if (strncasecmp(&hejstack[t],needle,neelen)==0)
    { return &hejstack[t]; }
  }

  return (char *)0;
}

void destroy()
{

#ifdef DEBUG
  if (debug==1)
  { printf("Mandel Server terminated.\n"); }
#endif

  fflush(fp);
  fclose(fp);

  exit(0);
}

#ifndef WINDOWS
void broken_pipe()
{
#ifdef DEBUG
  if (debug==1) printf("Signal thrown\n");
#endif

  setsignals();
}

void setsignals()
{
  signal(SIGPIPE,broken_pipe);
  signal(SIGURG,broken_pipe);
  signal(SIGIO,broken_pipe);
  signal(SIGHUP,broken_pipe);

  signal(SIGINT,destroy);
  signal(SIGTERM,destroy);
}
#endif

void read_config()
{
FILE *fp;
int r=0,w=0,t;
unsigned char name[100];

  fp=fopen(configdir,"r");
 
  if (fp==0)
  { 
    printf("No config file found.  Going with defaults.\n");
  }
    else
  {
    printf("Config File: %s\n",configdir);
    while(fgets(name,100,fp)!=NULL)
    {
      t=0;
      while((name[t]!=' ' && name[t]!='\t') && (name[t]!='\n' && name[t]!='\r'))
      { 
        if (t==100) break;
        t++; 
      }

      if (name[0]==0 || (name[0]=='\n' || name[0]=='\r')) name[0]='#';
      name[t++]=0;
      if (name[0]!='#')
      {
        w=0;
        if (strcasecmp(name,"REAL_START")==0) w=1;
          else
        if (strcasecmp(name,"REAL_END")==0) w=2;
          else
        if (strcasecmp(name,"IMAGINARY_START")==0) w=3;
          else
        if (strcasecmp(name,"IMAGINARY_END")==0) w=4;
          else
        if (strcasecmp(name,"COUNT")==0) w=5;
          else
        if (strcasecmp(name,"WIDTH")==0) w=6;
          else
        if (strcasecmp(name,"HEIGHT")==0) w=7;
          else
        if (strcasecmp(name,"FRAMES")==0) w=9;
          else
        if (strcasecmp(name,"PORT")==0) w=10;
          else
        if (strcasecmp(name,"MAXCONN")==0) w=11;
          else
        if (strcasecmp(name,"MINCONN")==0) w=12;
          else
        if (strcasecmp(name,"MAX_IDLE_TIME")==0) w=13;
          else
        if (strcasecmp(name,"REAL_START_LAST")==0) w=14;
          else
        if (strcasecmp(name,"REAL_END_LAST")==0) w=15;
          else
        if (strcasecmp(name,"IMAGINARY_START_LAST")==0) w=16;
          else
        if (strcasecmp(name,"IMAGINARY_END_LAST")==0) w=17;

        if (w!=0)
	{ 
          while(name[t]==' ' || name[t]=='\t')
	  {
	    if (name[t]==0 || (name[t]=='\r' || name[t]=='\n')) break;
	    t++;
	  }
	  r=t;
          while((name[t]!=' ' && name[t]!='\t') && ((name[t]!='\n' && name[t]!='\r') && name[t]!=0))
          { t++; }
	  if (r!=t)
	  {
	    name[t]=0;

	    if (w==1) mandelbrot.real_start=atof(&name[r]);
              else
	    if (w==2) mandelbrot.real_end=atof(&name[r]);
              else
	    if (w==3) mandelbrot.imaginary_start=atof(&name[r]);
              else
	    if (w==4) mandelbrot.imaginary_end=atof(&name[r]);
              else
	    if (w==5) mandelbrot.count=atoi(&name[r]);
              else
	    if (w==6) mandelbrot.width=atoi(&name[r]);
              else
	    if (w==7) mandelbrot.height=atoi(&name[r]);
              else
	    if (w==8) mandelbrot.zoom_length=atof(&name[r]);
              else
	    if (w==9) mandelbrot.frames=atoi(&name[r]);
              else
	    if (w==10) port=atoi(&name[r]);
	      else
	    if (w==11) maxconn=atoi(&name[r]);
	      else
	    if (w==12) minconn=atoi(&name[r]);
	      else
	    if (w==13) max_idle_time=atoi(&name[r])*60;
              else
	    if (w==14) mandelbrot.real_start_last=atof(&name[r]);
              else
	    if (w==15) mandelbrot.real_end_last=atof(&name[r]);
              else
	    if (w==16) mandelbrot.imaginary_start_last=atof(&name[r]);
              else
	    if (w==17) mandelbrot.imaginary_end_last=atof(&name[r]);
	  }
	}
      }
    } 
    fclose(fp);
  }
}

void message(int ID, char *daMessage)
{
/* unsigned char newMessage[2048]; */

  if (users[ID]->inuse!=1) return;
  { 
    if (send(users[ID]->socketid,daMessage,strlen(daMessage),0)<=0)
    {
      users[ID]->inuse=2;
      kill_user(ID);
    }
  }
}

int getNum(char *myString,int ID)
{
int r,len,out=0;

  r=0;
  len=strlen(myString);
  while (r<len && (myString[r]>='0' && myString[r]<='9'))
  {
    if (r<8) out=out*10+(myString[r]-48);
    r++;
  }

  if (r==0) 
  {
    if (ID!=-1)
    { message(ID,">> Unrecognized command"); }
    return -1;
  }

  if (ID!=-1)
  {
    return -1;
  }

  return out;
}

int strip(char *myString)
{
int r=0;

   while(myString[r]>='0' && myString[r]<='9')
   {
     /* if (myString[r]==0) return r; */
     r++;
   }

   while(myString[r]==' ')
   {
     /*  if (myString[r]==0) return r; */
     r++;
   }
   return r;
}

void timediff(int time1,int time2, char *thisString)
{
int temp;
unsigned char tempString[70];

  thisString[0]=0;
  if (time1<time2) { temp=time1; time1=time2; time2=temp; }
  temp=time1-time2;
  if (temp!=0) 
  { sprintf(tempString,"%dsec",temp%60); temp=temp/60; strcpy(thisString,tempString); }
  if (temp!=0)
  { sprintf(tempString,"%dmin %s",temp%60,thisString); temp=temp/60; strcpy(thisString,tempString); }
  if (temp!=0)
  { sprintf(tempString,"%dhours %s",temp%24,thisString); temp=temp/24; strcpy(thisString,tempString); }
  if (temp!=0)
  { sprintf(tempString,"%ddays %s",temp%365,thisString); temp=temp/365; strcpy(thisString,tempString); }
  if (temp!=0)
  { sprintf(tempString,"%dyears %s",temp,thisString); strcpy(thisString,tempString); }
}

void shorttimediff(int time1,int time2, char *thisString)
{
int temp;

  thisString[0]=0;
  if (time1<time2) { temp=time1; time1=time2; time2=temp; }
  temp=time1-time2;
  if (temp!=0) { sprintf(thisString,"%ds",temp%60); temp=temp/60; }
  if (temp!=0) { sprintf(thisString,"%dm",temp%60); temp=temp/60; }
  if (temp!=0) { sprintf(thisString,"%dh",temp%24); temp=temp/24; }
  if (temp!=0) { sprintf(thisString,"%dd",temp%365); temp=temp/365; }
  if (temp!=0) { sprintf(thisString,"%dy",temp); }
}

int ConvNum(char *myString)
{
int r=0;
unsigned char tempString[40];

  safestrncpy(tempString,myString,39);
  while(r<39)
  {
    if (tempString[r]<'0' || tempString[r]>'9') break;
    r++;
  }
  tempString[r]=0;
  if (r==0) return -1;

  return atoi(tempString);
}

int is_number(char *myString)
{
int r=0;

  while((myString[r]!=' ' && myString[r]!='\t') &&
	  ((myString[r]!='\n' && myString[r]!='\r') && myString[r]!=0))
  {
    if (myString[r]<'0' || myString[r]>'9') return 0;
    r++;
  }

  if (r==0) return 0;
  return 1;
}

void setup(int argc, char *argv[])
{
int r;

  port=DEFAULT_PORT;
  configdir[0]=0;
  debug=1;

  mandelbrot.real_start=-2.00;
  mandelbrot.real_end=1.00;
  mandelbrot.imaginary_start=-1.00;
  mandelbrot.imaginary_end=1.00;
  mandelbrot.count=100;
  mandelbrot.width=640;
  mandelbrot.height=480;
  mandelbrot.zoom_length=0.01;
  mandelbrot.frames=200;
  mandelbrot.real_start_last=-2.00;
  mandelbrot.real_end_last=1.00;
  mandelbrot.imaginary_start_last=-1.00;
  mandelbrot.imaginary_end_last=1.00;

  for (r=1; r<argc; r++)
  {
    if (strcasecmp(argv[r],"-f")==0) 
    { strcpy(configdir,argv[++r]); break; }
  }

  if (configdir[0]==0)
  { strcpy(configdir,"mandel_server.conf"); }

  uptime=time(NULL);

  read_config();

  for (r=1; r<argc; r++)
  {
    if (strcmp(argv[r],"-p")==0)
    { port=atoi(argv[++r]); }
      else
    if (strcasecmp(argv[r],"-d")==0) debug=1;
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("     Real start: %f\n",mandelbrot.real_start);
    printf("       Real end: %f\n",mandelbrot.real_end);
    printf("Imaginary start: %f\n",mandelbrot.imaginary_start);
    printf("  Imaginary end: %f\n",mandelbrot.imaginary_end);
    printf("          Width: %d\n",mandelbrot.width);
    printf("         Height: %d\n",mandelbrot.height);
    printf("          Count: %d\n",mandelbrot.count);
    printf("         Frames: %d\n",mandelbrot.frames);
    printf("\n");
  }
#endif

  mandelbrot.rs_step=(mandelbrot.real_start_last-mandelbrot.real_start)/mandelbrot.frames;
  mandelbrot.re_step=(mandelbrot.real_end_last-mandelbrot.real_end)/mandelbrot.frames;
  mandelbrot.is_step=(mandelbrot.imaginary_start_last-mandelbrot.imaginary_start)/mandelbrot.frames;
  mandelbrot.ie_step=(mandelbrot.imaginary_end_last-mandelbrot.imaginary_end)/mandelbrot.frames;

}

int read_param(FILE *fp, char *s)
{
int ch;
int r;

  r=0;
  s[0]=0;

  while(1)
  {
    ch=getc(fp);
    if (ch==EOF && r==0) return -1;
    if ((ch=='\n' || ch=='\r') && r==0) continue;
    if (ch=='\\') { s[r++]=getc(fp); continue; }
    if (ch==':' || ch=='\n' || ch=='\r' || ch==EOF) break;
    s[r++]=ch;
  }

  s[r]=0;
  if (ch==':') return 1;

  return 0;
}





