#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#ifndef WINDOWS
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <pthread.h>
#endif
#include <errno.h>
#include <time.h>
#include <string.h>
#ifdef WINDOWS
#include <windows.h>
#include <process.h>
#include <winsock.h>
#endif

#include "globals.h"
#include "functions.h"

/*

This code is copyright 2004-2005 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

*/

int sockfd,port;
struct User **users;
struct User nulluser;
struct mandelbrot_t mandelbrot;
int debug;
unsigned int uptime;
unsigned short int serverflags;
char configdir[256];
char dat_file[256]={ "frames/frames.dat" };
char welcome[256];
int welcomelen;
int maxconn;
int minconn;
int max_idle_time=0;
FILE *fp;

int init(int socketid, struct sockaddr_in *cli_addr, int clilen)
{
int k,r,ID;
struct hostent *my_hostent=NULL;
unsigned char *address=0;
unsigned char *fullmessage={ "Mandel Server is full\r\n" };

  if ((serverflags&2)==2)
  { my_hostent=gethostbyaddr((char *)&(cli_addr->sin_addr.s_addr),4,AF_INET); }

  for (r=0; r<maxconn; r++)
  { 
    if (users[r]->inuse==0) 
    { break; } 
  }

  if (r>=maxconn)
  {
#ifdef DEBUG
    if (debug==1) printf("Chat server full\n");
#endif
    send(socketid,fullmessage,strlen(fullmessage),0); 
    socketdie(socketid);
    return -1;
  }

  if (users[r]->idletime==-1)
  { 
#ifdef DEBUG
    if (debug==1) printf("Allocing new user %d\n",r);
#endif
    users[r]=malloc(sizeof(struct User)); 
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("searched for open User: found %d\n",r);
    fflush(stdout);
  }
#endif

  if (my_hostent!=NULL && (serverflags&2)==2)
  { safestrncpy(users[r]->location,my_hostent->h_name,61); }
    else
  { 
    address=(char*)&cli_addr->sin_addr.s_addr;
    sprintf(users[r]->location,"%d.%d.%d.%d",address[0],address[1],address[2],address[3]);
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("Connected from location: %s\n",users[r]->location);
    fflush(stdout);
  }
#endif

  k=0;
  if ((serverflags&8)==8)
  {
    for (k=0; k<maxconn; k++)
    {
      if (users[k]->inuse!=1) continue;
      if (strcmp(users[r]->location,users[k]->location)==0)
      { k=-1; break; }
    }
  }

  users[r]->ID=r;
  users[r]->socketid=socketid;
  ID=r;

  users[ID]->stamptime=0;
  users[ID]->ptr=0;
  users[ID]->lptr=0;
  users[ID]->lr=0;
  users[ID]->logontime=time(NULL);
  users[ID]->idletime=time(NULL);
  users[ID]->curr_frame=-1;
  users[ID]->upload_len=0;

  users[ID]->socketfd=fdopen(socketid,"rb+");

  if (users[ID]->socketfd==0) printf("Problem opening file desc.\n\n");

#ifndef WINDOWS
  fcntl(users[ID]->socketid, F_SETFL, O_NONBLOCK);
#else
  u_long arg=1L;
  ioctlsocket(users[ID]->socketid, FIONBIO, (u_long FAR *)&arg);
#endif

/* hmmm 

  if (setsockopt(sockfd,SOL_SOCKET,SO_RCVLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_RCVLOWAT can't be set.\n"); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_SNDLOWAT,&sopt,sizeof(sopt)))
  { printf("socket options SO_SNDLOWAT can't be set.\n"); }
*/
 
  send(users[ID]->socketid,welcome,welcomelen,0);

  users[ID]->inuse=1;

/*
  sprintf(tempString,YOU_LOGIN,ID,users[ID]->location);
  message(ID,tempString);
*/

  return 0;
}

void kill_user(int ID)
{
  if (users[ID]->curr_frame!=-1)
  {
    fseek(fp,users[ID]->curr_frame,SEEK_SET);
    putc(255,fp);
  }

#ifdef DEBUG
  if (debug==1)
  {
    printf("Getting ready to close socket.\n");
    if (users[ID]->inuse!=4)
      printf("Socket Close Return value: %d:\n",socketdie(users[ID]->socketid));
  }
    else
#endif
  { if (users[ID]->inuse!=4) socketdie(users[ID]->socketid); }

#ifdef DEBUG
  if (debug==1) printf("Connection dead: %d\n",ID);
  fflush(stdout);
#endif

  users[ID]->inuse=0;
  users[ID]->idletime=time(NULL);
}

int buffered_read(int ID)
{
int r,t=0; 

  while(1)
  {
    errno=0;
    if (users[ID]->lr>=users[ID]->lptr)
    {
      users[ID]->lptr=recv(users[ID]->socketid,users[ID]->myString,512-users[ID]->ptr,0);
      users[ID]->lr=0;
      if (users[ID]->lptr>=0) users[ID]->myString[users[ID]->lptr]=0;

      if (users[ID]->lptr==0) return -1;
      if (users[ID]->lptr<0)
      {
/*
        if (users[ID]->inuse==1)
         if (errno==EINTR || errno==EAGAIN) 
         { errno=0; return 0; }
 */ 
#ifdef DEBUG
        if (debug==1)
        { printf("Lost Connection (NULL): %d\n",ID); fflush(stdout); }
#endif

        users[ID]->ptr=0; 
        users[ID]->lptr=0;
        return -1;
      }
    }
  
    for (r=users[ID]->lr; r<users[ID]->lptr; r++)
    { 
      if ((users[ID]->myString[r]=='\n' || users[ID]->myString[r]=='\r') || users[ID]->myString[r]==0)
      { 
/*
        if (t!=0 || users[ID]->myString[r]==0)
        { t=-1; break; }
*/
        t=-1;
        break;
      }

      if (users[ID]->myString[r]==127 || users[ID]->myString[r]==8) 
      { if (users[ID]->ptr>0) users[ID]->ptr--; }
        else
      { if (users[ID]->myString[r]>=32 && users[ID]->myString[r]<255) users[ID]->outString[users[ID]->ptr++]=users[ID]->myString[r]; }

      t=1;
    }

    users[ID]->lr=r+1;

    if ((t==-1 && users[ID]->ptr>0) || users[ID]->ptr>512) 
    { 
      users[ID]->outString[users[ID]->ptr]=0;
      users[ID]->ptr=0;
      if (users[ID]->myString[users[ID]->lr]=='\n' || 
          users[ID]->myString[users[ID]->lr]=='\r')
      { users[ID]->lr++; }

      return 1; 
    }
      else
    { return 0; }
  }
}

void chat(int *params)
{
int t=0,r,trim,k;
unsigned char tempString[1500];
int ID=0;
int msock=0;
fd_set readset;
struct timeval tv;
char *outString;
int gc_time,thread_num;
int dirty_buffer;
char *command,*param;
double real_start,real_end,imaginary_start,imaginary_end;

  thread_num=params[0];
  gc_time=time(NULL);

  while(1)
  {
    if (time(NULL)-gc_time>GC_TIME)
    { 
      for (r=minconn+thread_num; r<maxconn; r=r+MAX_USER_THREADS)
      {
        if (users[r]->inuse==0 && users[r]->idletime!=-1)
        {
#ifdef DEBUG
          if (debug==1) printf("Deallocating line %d\n",r);
#endif
          if (time(NULL)-users[r]->idletime>GC_TIME)
          free(users[r]);
          users[r]=&nulluser;
        }
      }

/*
      if (max_idle_time>0)
      {
        for (r=thread_num; r<maxconn; r=r+MAX_USER_THREADS)
        {
          if ((users[r]->inuse!=0 && users[r]->user_level<5) && time(NULL)-users[r]->idletime>max_idle_time)
          {
	    message(r,">> Too long without input.  Bye.");
	    users[r]->inuse=2;
            kill_user(r);
          }
        }
      }
*/
      gc_time=time(NULL);
    }

/* printf("HERE AGAIN\n"); */

    FD_ZERO(&readset);
/*
    FD_SET(sockfd,&readset);
    msock=sockfd;
*/
    msock=0;
    dirty_buffer=0;

    for (r=thread_num; r<maxconn; r=r+MAX_USER_THREADS)
    {
      if (users[r]->inuse==1)
      {
        FD_SET(users[r]->socketid,&readset);
        if (msock<users[r]->socketid) msock=users[r]->socketid;
        if (users[r]->lr<users[r]->lptr) dirty_buffer=1;
      }
    }

    if (dirty_buffer==0)
    {    
      tv.tv_sec=1;
      tv.tv_usec=0;
    }
      else
    {
      tv.tv_sec=0;
      tv.tv_usec=1;
    }

#ifdef WINDOWS
    if (msock==0)
    {
      _sleep(1000);
      continue;
    }
#endif

    if ((t=select(msock+1,&readset,NULL,NULL,&tv))==-1)
    {
#ifdef WINDOWS
      if (WSAGetLastError()!=WSANOTINITIALISED) printf("yes %d\n",errno);
      if (WSAGetLastError()!=WSAEINTR)
#else
      if (errno!=EINTR)
#endif
      { 
#ifdef DEBUG
        if (debug==1) printf("not EINTR %d\n",thread_num);
#endif
      }
         else
      { 
#ifdef DEBUG
        if (debug==1) printf("Problem with select\n"); 
#endif
      }
    }

/*
printf("Heartbeat %d  --- %d\n",thread_num,t);
*/

    for (ID=thread_num; ID<maxconn; ID=ID+MAX_USER_THREADS)
    { 
      errno=0;
      if (users[ID]==0) continue;
      if (users[ID]->inuse!=1) continue;

/*
printf("Checking line %d     %d %d %d\n",ID,users[ID]->lr,users[ID]->lptr,users[ID]->ptr);
*/

      if (users[ID]->lr<users[ID]->lptr)
      {
      /* printf("dirty\n"); */
      }
        else
      if (FD_ISSET(users[ID]->socketid,&readset)) 
      {  } 
        else 
      continue;

      if (users[ID]->upload_len==-1)
      {
        t=recv(users[ID]->socketid,tempString,4,0);
        users[ID]->upload_len=(tempString[0]<<24)+(tempString[1]<<16)+(tempString[2]<<8)+tempString[3];

#ifdef DEBUG
        if (debug==1)
        {
          printf("Line %d is uploading a file of size %d\n",ID,users[ID]->upload_len);
        }
#endif
        continue;
      }
        else
      if (users[ID]->upload_len!=0)
      {
        t=users[ID]->upload_len;
        if (t>1024) t=1024;
        r=recv(users[ID]->socketid,tempString,t,0);
        users[ID]->upload_len=users[ID]->upload_len-r;
        k=0;
        while(k!=r)
        {
          k=k+fwrite(tempString+k,1,r-k,users[ID]->out);
        }

/*
#ifdef DEBUG
        if (debug==1)
        {
          printf("%d uploading %d bytes left\n",ID,users[ID]->upload_len);
        }
#endif
*/

        if (users[ID]->upload_len==0)
        {
          fclose(users[ID]->out);

          fseek(fp,users[ID]->curr_frame,SEEK_SET);
          putc(254,fp);
          fflush(fp);
          users[ID]->curr_frame=-1;
        }
        continue;
      }
        else
      {
        r=buffered_read(ID);
      }

      if (r<0) kill_user(ID);
      if (r<1) continue;

/*
printf("FILE: %d\n",ID);
fflush(stdout);

printf("READ: %s\n\n",users[ID]->outString);
fflush(stdout);
*/

      outString=users[ID]->outString;

      trim=0;

      while(outString[trim]==' ') trim++;

      users[ID]->idletime=time(NULL);

#ifdef DEBUG
      if (debug==1)
      {
        printf("Read in: %d bytes on thread %d.\n",(int)strlen(outString),thread_num);
        printf("%d typed: %s\n",ID,outString);
        fflush(stdout); 
      }
#endif

      r=0;
      while(outString[r]==' ' && outString[r]!=0) r++;

      command=outString+r;
      r=0;
      while(command[r]!=' ' && command[r]!='\r' && command[r]!='\n' && command[r]!=0) r++;
      command[r++]=0;

      #ifdef DEBUG
      if (debug==1)
      {
        printf("command=%s\n",command);
      }
      #endif

      param=command+r;
      r=strlen(param)-1;
      while((param[r]==' ' || param[r]=='\r') || param[r]=='\n') param[r--]=0;

      if (strcasecmp(command,"quit")==0)
      {
        message(ID, ">> Bye Bye\r\n");
        kill_user(ID);
      }
        else
      if (strcasecmp(command,"info")==0)
      {
        timediff(time(NULL),uptime,outString);
        sprintf(tempString,">> Up for %s\r\n",outString);
        message(ID,tempString);
        sprintf(tempString,">> At the tone, the time will be %s\r\n",ctime((time_t*)&users[ID]->idletime));
        message(ID,tempString);
      }
        else
      if (strcasecmp(command,"next")==0)
      {
        if (users[ID]->curr_frame!=-1)
        {
          message(ID,">> Error: You already have a frame\r\n");
        }
          else
        {
          fseek(fp,0,SEEK_SET);
          while((r=getc(fp))!=EOF)
          { if (r==255) break; }

          if (r==EOF)
          {
            message(ID,">> Done\r\n");
            continue;
          }
          k=ftell(fp)-1;

          real_start=mandelbrot.real_start+(k*mandelbrot.rs_step);
          real_end=mandelbrot.real_end+(k*mandelbrot.re_step);
          imaginary_start=mandelbrot.imaginary_start+(k*mandelbrot.is_step);
          imaginary_end=mandelbrot.imaginary_end+(k*mandelbrot.ie_step);

          sprintf(outString,">> compute %d,%d,%f,%f,%f,%f,%d,%d\r\n",mandelbrot.width,mandelbrot.height,real_start,real_end,imaginary_start,imaginary_end,mandelbrot.count,k);
          message(ID,outString);
          fseek(fp,k,SEEK_SET);
          putc(ID,fp);
          fflush(fp);
          users[ID]->curr_frame=k;
        }
      }
        else
      if (strcasecmp(command,"upload")==0)
      {
        sprintf(tempString,"frames/frame%05d.gif",users[ID]->curr_frame);
        users[ID]->out=fopen(tempString,"wb");
        if (users[ID]->out==0)
        {
          printf("Problem opening file %s for writing\n",tempString);
          message(ID,">> Error: server file access error\r\n");
          continue;
        }

        message(ID,">> Send\r\n");
        users[ID]->upload_len=-1;


      }
        else
      { message(ID,">> Unrecognized command\r\n"); }
    }
  }
}

int main(int argc,char *argv[])
{
struct sockaddr_in serv_addr;
int r,ch,sopt=1;
struct linger mylinger;
int newsockfd,clilen;
struct sockaddr_in cli_addr;
int threads[MAX_USER_THREADS];
#ifndef WINDOWS
pthread_t pid;
#endif
#ifdef WINDOWS
WSADATA wsaData;
WORD wVersionRequested=MAKEWORD(1,1);
int Win32isStupid;

  Win32isStupid=WSAStartup(wVersionRequested, &wsaData);
  if (Win32isStupid)  /* Shouldn't this always be true? :) */
  {
    printf("Winsock can't start.\n");
    exit(1);
  }

/*
  printf("Max sockets: %d\n",wsaData.iMaxSockets);
  printf("Description: %s\n",wsaData.szDescription);
*/
#endif

  maxconn=MAXCONN;
  minconn=10;
  max_idle_time=0;
  setup(argc,argv);

#ifndef WINDOWS
  if (debug==0)
  {
    r=fork();

    if (r==-1)
    { perror("Error: Could not fork\n"); exit(3); }
      else
    if (r==0)
    {
      close(STDIN_FILENO);
      close(STDERR_FILENO);

      if (setsid()==-1)
      { exit(4); }
    }
      else
    { return 0; }
  }
#endif

#ifdef DEBUG
  if (debug==1)
  {
    printf("minconn: %d\n",minconn);
    printf("maxconn: %d\n",maxconn);
  }
#endif

  if ((sockfd=socket(AF_INET,SOCK_STREAM,0))<0)
  { printf("Can't open socket.\n"); exit(1); }

  /* bzero((char*)&serv_addr,sizeof(serv_addr)); */
  memset((char*)&serv_addr,0,sizeof(serv_addr));
  serv_addr.sin_family=AF_INET;
  serv_addr.sin_addr.s_addr=htonl(INADDR_ANY);
  serv_addr.sin_port=htons(port);

  mylinger.l_onoff=0;
  mylinger.l_linger=1;

#ifndef WINDOWS
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&sopt,sizeof(sopt)))
  { printf("socket options REUSERADDR can't be set.\n"); exit(1); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,&mylinger,sizeof(struct linger)))
  { printf("socket options LINGER can't be set.\n"); exit(1); }
#else
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,(char *)&sopt,sizeof(sopt)))
  { printf("socket options REUSERADDR can't be set.\n"); exit(1); }
  if (setsockopt(sockfd,SOL_SOCKET,SO_LINGER,(char *)&mylinger,sizeof(struct linger)))
  { printf("socket options LINGER can't be set.\n"); exit(1); }
#endif


  if(bind(sockfd,(struct sockaddr *)&serv_addr,sizeof(serv_addr))<0)
  { printf("server can't bind.\n"); exit(1); }

  listen(sockfd,5);
#ifndef WINDOWS
  setsignals();
#endif

/* YOU MAY NOT EDIT THE NEXT LINE OF CODE */
  printf("\n"VERSION"\n"COPYRIGHT"\n\n");
/* NAKEN CHAT IS NOT GPL */

  printf("Started on port: %d\n",port);
  fflush(stdout);

  if (debug==0) close(STDOUT_FILENO);

  nulluser.inuse=0;
  nulluser.idletime=-1;
  users=malloc(sizeof(char *)*maxconn);

  for (r=0; r<maxconn; r++)
  {
    users[r]=&nulluser;
  }

  for (r=0; r<minconn; r++)
  {
    users[r]=malloc(sizeof(struct User));
    users[r]->inuse=0;
  }

  for (r=0; r<MAX_USER_THREADS; r++)
  {
    threads[r]=r;
#ifndef WINDOWS
    pthread_create(&pid,NULL,(void*)chat,&threads[r]);
#else
    _beginthread((void*)chat,0,&threads[r]);
#endif
  }

  /* chat(); */

  fp=fopen(dat_file,"r");
  if (fp==0)
  {
    fp=fopen(dat_file,"wb");
    if (fp==0)
    {
      printf("Error opening %s for writing\n",dat_file);
      exit(1);
    }
    for (r=0; r<mandelbrot.frames; r++) putc(255,fp);
  }
  fclose(fp);

  fp=fopen(dat_file,"rb+");
  if (fp==0)
  {
    printf("Error opening %s for appending\n",dat_file);
    exit(1);
  }

  fseek(fp,0,SEEK_SET);
  while((ch=getc(fp))!=EOF)
  {
    if (ch<254)
    {
      fseek(fp,ftell(fp)-1,SEEK_SET);
      putc(255,fp);
    }
  }
  fflush(fp);

  strcpy(welcome,">> "VERSION"\n>> "COPYRIGHT"\n");
  welcomelen=strlen(welcome);

  clilen=sizeof(cli_addr);

  while(1)
  {
    newsockfd=accept(sockfd,(struct sockaddr *)&cli_addr,&clilen);

    if (newsockfd<0)
    {
#ifdef DEBUG
      if (debug==1) { printf("server accept error.\n"); }
#endif
    }
      else
    { 
#ifdef DEBUG
      if (debug==1)
      {
        printf("New socket accepted\n");
        fflush(stdout);
      }
#endif
      init(newsockfd,&cli_addr,clilen);
    }
  }

  fclose(fp);

  return 0;
}

